#ifndef PCADLLH_H
#define PCADLLH_H

#define WM_PUTINFO	WM_USER + 100

/*#ifndef __cplusplus
extern "C" {
#endif*/ /*__cplusplus*/

void GiveProcess();
void Msg(int level, char *format, ...);
void Fail(int err);
void ByteToStr(char *str, long byte);
long lmin(long i1, long i2);

/*#ifndef __cplusplus
}
#endif*/ /*__cplusplus*/

#endif /*PCADLLH_H*/

/*--------------------------------------------------------------------------------------------*/

#ifndef MYEXCPT_H
#define MYEXCPT_H

class CMyException : public CException
	{
	public:
		CMyException();
		CMyException(int i);
		
		int err;
	};

#endif /*MYEXCPT_H*/

/*--------------------------------------------------------------------------------------------*/

#ifndef DEF_H
#define DEF_H

#define XSPACE		5
#define YSPACE		5
#define YFACETEXT	3
#define YTEXT		12
#define MAXXFACE	10
#define MAXFLOAT 	30000

/* Commands */
#define CMD_TRAINING							0
#define CMD_RECONSTRUCTION					1
#define CMD_RECOGNIZE						2
#define CMD_MEET								3
#define CMD_LOAD_EIGEN						4
#define CMD_SAVE_EIGEN_DB					5
#define CMD_SAVE_EIGEN						6
#define CMD_SAVE_REPORT						7
#define CMD_SAVE_CMP_MEET					8
#define CMD_SAVE_CMP_RECONSTRUCTION		9
#define CMD_SAVE_CMP_RECOGNITION			10
#define CMD_MAX								11

typedef struct _GEOMPARAMS
	{
	int xNum;						/*Number of columns of faces */
	int yNum;						/*Number of lines of faces -1*/
	int lastNum;					/*Number of columns of faces on the last line*/
	}GEOMPARAMS;

typedef struct _CORNER
	{
	int x, y;
	}CORNER;

#endif /* DEF_H */

/*--------------------------------------------------------------------------------------------*/

#ifndef MEMORY_H
#define MEMORY_H

class CMemory	
	{
	public:
		CMemory();
		~CMemory();
		
		void FAR	*GetPtr();
		int 		ReleasePtr();
		int 		Allocate(long lsizet);
		int 		Reallocate(long lsizet);
		
		long		GetSize();
		HGLOBAL	GetHandle();
		
	private:
		HGLOBAL 	handle;
		long 		lsize;
	};

#endif /*MEMORY_H*/

/*--------------------------------------------------------------------------------------------*/

#ifndef MATRIX_H
#define MATRIX_H

#define		valtype	float
#define		valtype2	float
#define 		MAX_NAME 10

class CMatrixSym;
class CVect;
class CDiag;

class CMatrix : public CObject
	{
	protected:
		long col, li;         						/*Number of columns and lines of the matrix*/
		CMemory *pMem;								/*Ptr to a CMemory obj containing the*/
													/*values : a(1,1) a(1,2) ... a(1,col) ... a(li,col)*/
		valtype *pData;								/*Pointer to the matrix's value if pMem is locked*/
		BOOL bTranspose;							/*tells if the matrix is transposed*/
		BOOL bLocked;								/*tells if pMem is locked */
      char **name;											/*points to a list of string, containing the name of each column vector*/

	public:
		CMatrix();
		CMatrix(long l, long c);
		~CMatrix();
		int Save(char *filename);
		int Load(char *filename);
		void Add(CMatrix *pMat2);					/*Add to this matrix the matrix pMat2*/
		void Multiply(CMatrix *pMat1, CMatrix *pMat2, BOOL bResize = FALSE, long lNewRow = 0, long lNewCol = 0);
													/*Put into this pMat1*pMat2*/
		void Multiply(CDiag *pDiag, BOOL bResize = TRUE);
													/*Multiply this matrix by the diagonal matrix pDiag*/
		void NormalizeCol();						/*Normalize each column*/
		void OuterProduct(CMatrixSym *pMat);		/*Compute the outer product of pMat*/
		void InnerProduct(CMatrixSym *pMat);		/*Compute the inner product of pMat*/
		void InnerProduct(CMatrix *pMat);			/*Compute the inner product of pMat*/
		void Initialize();							/*Sets all the elements to 0*/
		void SetTranspose(BOOL bTransposet);		/*Set the bTranspose parameter*/
		virtual long GetCol();
		virtual long GetLi();
		virtual void Lock();						/*Lock pMem and validate pData*/
		virtual void Unlock();						/*Unlock pMem and Unvalidate pData*/
		virtual int Eigen(CVect *ev, long leigen);	/*Compute the eigenvectors of this matrix*/
		void Expand(long c, float *scale, float *shift, valtype dest_min, valtype dest_max);
		
		int GetColName(long c, char *str, int len);	/*copy in str the name of the column vector c*/
		int SetColName(long c, char *str);				/*set the name of the column vector c to str*/
		
	public:
		virtual void Allocate(long, long);
		virtual void Allocate();					/*Allocate (or reallocate) the memory*/
													/*needed by pData; Call the CMemory's Allocate fct*/
	public:
		virtual valtype GetAt(long l, long c);
		virtual void SetAt(long l, long c, valtype value);
	};
	
#endif /*MATRIX_H*/

/*--------------------------------------------------------------------------------------------*/

#ifndef MATRIXPTR_H
#define MATRIXPTR_H

class CMatrixPtr : public CMatrix
	{
	private:
		CMemory **pColPtr;							/*Locked pointer of CMatrix's pMem*/
													/*pMem contain the CMemory pointer to the columns of the matrix*/
		CMemory *pMemPtr;							/*Array of memory containing the pointers to the columns' values*/
		valtype2 **pData2;								/*Locked pointer of pMemPtr. So it contains the pointers to the values*/
		BOOL bAlloc;
		
	public:
		CMatrixPtr();
		CMatrixPtr(long l, long c);
		~CMatrixPtr();
		virtual valtype2 GetAt(long l, long c);			/*Get the element at (l,c) of the matrix*/
													/*if bTranspose == FALSE, and of the*/
													/*transposed matrix if bTranspose == TRUE*/
		virtual void SetAt(long l, long c, valtype2 value);
													/*Obvious. Becarful to bTranspose*/
		virtual void Lock();
		virtual void Unlock();
		void SetColumn(long c, CMemory *pCol);
													/*Assign the column c of the matrix to the memory vector pCol*/
		CMemory *GetColumn(long c);					/*Return the column c of the matrix as a CMemory object*/
		virtual void Allocate();					/*Allocate (or reallocate) the memory*/
													/*needed by pData; Call the CMemory's Allocate fct*/
		virtual void Allocate(long, long);			/*set li and col, call Allocate() and AllocateCol()*/
		void AllocateCol();							/*Allocate memory for the coulmns' data*/
		void FreeCol();								/*free the memory for the columns data*/
		void InvertCol();							/*Invert the order of this matrix*/
		void SwapCol(long l1, long l2);				/*Swap the order of two columns*/
		void Discard(long l);						/*Discard a column*/
	};

#endif /*MATRIXPTR_H*/

/*--------------------------------------------------------------------------------------------*/

#ifndef MATRXSYM_H
#define MATRXSYM_H

class CMatrixSym : public CMatrix
	{
	private:
	public:
		CMatrixSym();
		CMatrixSym(long l);
		~CMatrixSym();
		virtual long GetCol();
		virtual long GetLi();
		
		int Eigen(CMatrix *Q, CVect *ev);	/* Puts in Q the eigenvectors and in ev the eigenvalues of the matrix*/
		int Put(CMatrix *Q);				/* Puts all the elements of this matrix into the pointer to a CMatrix object*/
	private:
		virtual void Allocate();			/*Allocate (or reallocate) the memory*/
											/*needed by pData; Call the CMemory's Allocate fct*/
	public:
	virtual valtype GetAt(long l, long c);
	virtual void SetAt(long l, long c, valtype value);
	};

#endif /*MATRXSYM_H*/

/*--------------------------------------------------------------------------------------------*/

#ifndef CVECT_H
#define CVECT_H

class CVect
	{
	protected:
		long li;	         						/*Number of rows of the vector*/
		CMemory *pMem;								/*Ptr to a CMemory obj containing the*/
													/*values : a(1,1) a(1,2) ... a(1,col) ... a(li,col)*/
		valtype *pData;								/*Pointer to the matrix's value if pMem is locked*/
		BOOL bLocked;								/*tells if pMem is locked */
	public:
		CVect();
		CVect(long c);
		~CVect();
		int Save(char *);							/*Save the vector in a text file*/
		int Load(char *);							/*Load the vector from a text file*/
		void Initialize();							/*Sets all the elements to 0*/
		long GetLi();
		void Lock();								/*Lock pMem and validate pData*/
		void Unlock();								/*Unlock pMem and Unvalidate pData*/
		void PowCmp(float exp);						/*Pow each individual composant of the vector*/
		
		void Allocate();							/*Allocate (or reallocate) the memory*/
													/*needed by pData; Call the CMemory's Allocate fct*/
		void Invert();								/* Invert the order of the columns, but keep the matrix diagonal*/
		void SwapCol(long l1, long l2);				/*Swap two elements*/
		void Discard(long l);						/*Discard one element*/	
		valtype GetAt(long l)
		 	/*{
#ifdef CHECK_BOUNDARY
		 	if (l <= 0 || l > li)
		 	 	Fail(NULL);
#endif
			return pData[l-1L];
		 	}*/;
		void SetAt(long l, valtype value)
		 	/*{
#ifdef CHECK_BOUNDARY
		 	if (l <= 0 || l > li)
		 		Fail(NULL);
#endif		 
			pData[l-1L] = value;
		 	}*/; 
	};
	
#endif /* CVECT_H */

/*--------------------------------------------------------------------------------------------*/

#ifndef CDIAG_H
#define CDIAG_H

class CDiag : public CVect
	{
	protected:
		
	public:
		CDiag();
		CDiag(long c);
		~CDiag();
	
		void Inverse();
		void Pow(float exp);
		long GetCol();
		void Multiply(CMatrix *pMat1, CMatrix *pMat2);
	};
	
#endif /* CDIAG_H */

/*--------------------------------------------------------------------------------------------*/

#ifndef CFACE_H
#define CFACE_H

//#include <stdio.h>

#define ASCII		0
#define BINARY		1
#define PPM			0
#define PGM			1
#define typeface	float

class CFace	: public CObject	
	{
	public:
		CFace();
		CFace(long, long);
		CFace(char *tFileName, char *tFaceName);
		~CFace();
		
		int Load();
		int Save();
		int Copy(CMatrixPtr *pMat, long c);
		void Allocate();
		
		void Expand(float *scale, float *shift, float src_min, float src_max);
		
		void SetFileName(char *str_file);
		char *GetFileName();
		char *GetName();
		void SetName(char *str_file, int len = -1);
		
		long Getxlen();
		long Getylen();
		
		typeface *GetData();
		void ReleaseData();
		
		CMemory *GetMem();
		
		void SetAt(long x, long y, typeface val);
		void SetAt(long lpixel, typeface val);
		typeface GetAt(long x, long y);
		typeface GetAt(long lpixel);
		
		void Lock();
		void Unlock();

	private:
		long xlen, ylen;						/*Dimensions of the image*/
		CMemory *data;							/*Memory object representing the image's bytes*/
		typeface *pdata;						/*pointer to the locked CMemory*/
		char cFormat;							/*Format of the input file, can be either ASCII or BINARY*/
		char cFileType;							/*File format. Can be PPM or PGM*/
		char FileName[255];						/*File name*/
		char FaceName[64];						/*Face's name*/
		BOOL bMat;								/*FALSE if the data are read from the disk
												  TRUE is they are from a computed matrix*/
		CMatrixPtr *pMat;							/*if the face is a matrix's column, pMat hold the pointer to it*/
		long col;								/* and col hold the column's number*/
		BOOL bAlloc;							/*TRUE if the object allocate itself the memory (So it has to delete it)*/
		
		int ReadASCII(unsigned char *str);		/*read an ascii image file*/
		int ReadBinary(unsigned char *str);		/*read a binary image file*/
		char *GetNextTok(char *str);
		char *GetNextLine(char *str);
	};

#endif /* CFACE_H */

/*--------------------------------------------------------------------------------------------*/

#ifndef CFaceDB_H
#define CFaceDB_H

class CFaceDB
	{
	private:
		long 			nNumFace;							/*Number of faces in the data base*/
		long 			xlen, ylen;							/*dimension of each face*/
		CObArray 	*pArray;								/*Actual array of CFace objects, template of CPtrArray*/
	
	public:
		CFaceDB();
		~CFaceDB();
		
		int 			Load(char *file);
		int 			AddFace(char *file);
		int 			Save(char *str_dir);
		int 			Copy(CMatrixPtr *pMat, CDiag *pDiag);
		
		long 			Getxlen();
		long 			Getylen();
		long 			GetNumPixel();
		long 			GetNumFace();
		CFace 		*GetFace(int l);
		typeface 	*GetFaceData(int l);
		CMemory 		*GetFaceMem(int l);
		char 			*GetFaceName(int l);
		void 			ReleaseFaceData(int l);
		void 			Expand(int nFace, float *scale, float *shift, float src_min, float src_max);
		
	private:
		int 			GetStr(char *str_temp, char *aName, int lenaName, char *Delimit);
		long 			GetLong(char *str_temp);
		int 			GetLine(char *str_temp, char *FileName, int lenFileName, char *FaceName, int lenFaceName);
	}; 
	
#endif /* CFaceDB_H */

/*--------------------------------------------------------------------------------------------*/

#ifndef CDRAWF_H
#define CDRAWF_H

class CDrawFace : public CFrameWnd
	{
	private:
		CFaceDB 			*pFaceDB;							/*Face DataBase data*/
		CMatrix 			*pFaceMat;
		CDiag 			*pEigenVal;
		char 				bMat;
		GEOMPARAMS 		gp;
		
		int				nMaxVPos;
		int				nVPos;
			
		HFONT 		hfont, hfontOld;		/*Character font*/
		int 			MapModePrevious;

	public:
		CDrawFace();
		
		void				SetFaceDB(CFaceDB *pFaceDBt);
		void				SetFaceDB(CMatrix *pFaceMatt);
		void				SetFaceDB(CMatrix *pFaceMatt, CDiag *pEigenValt);
		void				GetGeomParams(GEOMPARAMS *gp);                       
		int				Getxlen();
		int				Getylen();
	
	private:
		void				GetCornerFace(CORNER *c, long nFace);
		void				GetCornerName(CORNER *c, long nFace);

	// Generated message map functions
	protected:
		//{{AFX_MSG(CWndFace)
		afx_msg void OnPaint();
		afx_msg void OnVScroll(UINT nSBCode, UINT nPos, CScrollBar* pScrollBar);
		afx_msg int OnCreate(LPCREATESTRUCT lpCreateStruct);
		afx_msg void OnSize(UINT nType, int cx, int cy);
		//}}AFX_MSG
		DECLARE_MESSAGE_MAP()
	};
#endif /* CDRAWF_H */

/*--------------------------------------------------------------------------------------------*/

#ifndef CDRAWREC_H
#define CDRAWREC_H

class CDrawReconstruction : public CFrameWnd
	{
	private:
		CFaceDB *pFaceDB1;								/*Face DataBase1 data*/
		CFaceDB *pFaceDB2;
		GEOMPARAMS gp;
		
		int nVPos;
		int nMaxVPos;
		int nHPos;
		int nMaxHPos;
			
		HFONT 		hfont, hfontOld;		/*Character font*/
		int 			MapModePrevious;

	public:
		CDrawReconstruction();
		
		void				SetFaceDB(CFaceDB *pFaceDB1t, CFaceDB *pFaceDB2t);
		void				GetGeomParams(GEOMPARAMS *gp);
		int				Getxlen();
		int				Getylen();

	private:
		void	GetCornerFace(CORNER *c, int nFace, int num);
		void	GetCornerName(CORNER *c, int nFace, int num);

	// Generated message map functions
	protected:
		//{{AFX_MSG(CWndFace)
		afx_msg void OnPaint();
		afx_msg void OnVScroll(UINT nSBCode, UINT nPos, CScrollBar* pScrollBar);
		afx_msg int OnCreate(LPCREATESTRUCT lpCreateStruct);
		afx_msg void OnSize(UINT nType, int cx, int cy);
		afx_msg void OnHScroll(UINT nSBCode, UINT nPos, CScrollBar* pScrollBar);
		//}}AFX_MSG
		DECLARE_MESSAGE_MAP()
	};

#endif /* CDRAWREC_H */

/*--------------------------------------------------------------------------------------------*/

#ifndef ERROR_H
#define ERROR_H
/* File */
#define ERR_FILE_READ					-2
#define ERR_FILE_OPEN					-3
/* File Format */
#define ERR_FF							-20
#define ERR_FF_TRAINING_ENDTAG			-21
#define ERR_FF_RECONSTRUCTION_ENDTAG	-22
#define ERR_FF_RECOGNIZE_ENDTAG			-23
#define ERR_FF_MEET_ENDTAG				-24
#define ERR_FF_LOAD_EIGEN_ENDTAG		-25
/* Memory */
#define ERR_MEM_GETPTR					-50
#define ERR_MEM_ALLOCATE				-51
#define ERR_MEM_REALLOCATE				-52
#define ERR_MEMORY						-53
/* Matrix */
#define	ERR_MAT_LOAD_ROW				-100
#define	ERR_MAT_LOAD_COL				-101
#define	ERR_MAT_LOAD_HDR				-103
#endif /* ERROR_H */

/*--------------------------------------------------------------------------------------------*/

#ifndef STR_H
#define STR_H

/* Useful strings */
extern char chr_RT1;
extern char chr_RT2;
extern char chr_BTAG;
extern char str_BTAG[];
extern char chr_PROJ_COMMENT;
extern char str_RET[];
extern char str_TABRT[];
extern char str_END[];
extern char str_DIR[];
extern char chr_DIR;
extern char str_EXT_TXT[];
extern char str_SAVE_NUM[];

/* face file opening strings */
extern char str_ERR_FILE_OPEN[];
extern char str_ERR_FILE_READPROB[];
extern char str_ERR_FILE_WRITE[];
extern char str_ERR_FILE_EXT[];
extern char str_ERR_FILE_LOCATE[];
extern char str_ERR_FILE_CLOSE[];
extern char str_ERR_ASCIIBIN[];
extern char str_ERR_FILE_UNEXPECTEDEOF[];
extern char str_ERR_NOT255[];
extern char str_LOADED[];
extern char str_LOADING[];
extern char str_SAVING[];
extern char str_SAVED[];

/* face database */
extern char str_ERR_DIMENSION[];

/* project file strings */
extern char str_TRAINING[];
extern char str_RECONSTRUCTION[];
extern char str_RECOGNIZE[];
extern char str_MEET[];
extern char str_SAVE_EIGEN[];
extern char str_SAVE_EIGEN_DB[];
extern char str_SAVE_REPORT[];
extern char str_SAVE_CMP_MEET[];
extern char str_SAVE_CMP_RECONSTRUCTION[];
extern char str_SAVE_CMP_RECOGNIZE[];
extern char str_LOAD_EIGEN[];
extern char str_PROJECT_TRAINING[];
extern char str_PROJECT_TRAINING_DONE[];
extern char str_PROJECT_RECONSTRUCTION[];
extern char str_PROJECT_RECONSTRUCTION_DONE[];
extern char str_PROJECT_RECOGNIZE[];
extern char str_PROJECT_RECOGNIZE_DONE[];
extern char str_PROJECT_MEET[];
extern char str_PROJECT_MEET_DONE[];
extern char str_PROJECT_LOAD_EIGEN[];
extern char str_PROJECT_LOAD_EIGEN_DONE[];
extern char str_PROJECT_SAVE_EIGEN[];
extern char str_PROJECT_SAVE_EIGEN_DONE[];
extern char str_PROJECT_SAVE_EIGEN_DB[];
extern char str_PROJECT_SAVE_EIGEN_DB_DONE[];
extern char str_PROJECT_SAVE_REPORT[];
extern char str_PROJECT_SAVE_REPORT_DONE[];
extern char str_PROJECT_SAVE_CMP_MEET[];
extern char str_PROJECT_SAVE_CMP_MEET_DONE[];
extern char str_PROJECT_SAVE_CMP_RECONSTRUCTION[];
extern char str_PROJECT_SAVE_CMP_RECONSTRUCTION_DONE[];
extern char str_PROJECT_SAVE_CMP_RECOGNITION[];
extern char str_PROJECT_SAVE_CMP_RECOGNITION_DONE[];
extern char str_EVAL_FILENAME[];
extern char str_ALREADY_LOAD_EIGEN[];
extern char str_ALREADY_TRAINING[];
extern char str_NOTYET_EIGEN[];
extern char str_NOTYET_MEET[];
extern char str_NOTYET_RECONSTRUCTION[];
extern char str_NOTYET_RECOGNITION[];
extern char str_ERR_WRONGARG[];
extern char str_ERR_TRAINING[];
extern char str_ERR_FILE_READ[];
extern char str_ERR_TRAINING_ENDTAG[];
extern char str_ERR_RECONSTRUCTION_ENDTAG[];
extern char str_ERR_RECOGNIZE_ENDTAG[];
extern char str_ERR_MEET_ENDTAG[];
extern char str_ERR_LOAD_EIGEN_ENDTAG[];
extern char str_PROJECT_LOADING[];
extern char str_ERR_MISMATCH[];
extern char str_ERR_AAMNOTCREATE[];

/* Memory */
extern char str_ERR_MEM_ALLOCATE[];
extern char str_ERR_MEM_REALLOCATE[];
extern char str_ERR_MEM_GETPTR[];

/* Matrix */
extern char str_TOOITER[];
extern char str_ERR_MAT_COPY[];
extern char str_ERR_MAT_LOAD_ROW[];
extern char str_ERR_MAT_LOAD_COL[];
extern char str_ERR_MAT_LOAD_HDR[];

/* Auto-Associative Memory */
extern char str_AAM_BUILD_START[];
extern char str_AAM_BUILD_NORMALIZED[];
extern char str_AAM_BUILD_INNER[];
extern char str_AAM_BUILD_EIGEN1[];
extern char str_AAM_BUILD_SORT[];
extern char str_AAM_BUILD_DONE[];
extern char str_AAM_RECONSTRUCTION_START[];
extern char str_AAM_RECONS_NORMALIZED[];
extern char str_AAM_RECONSTRUCTION_DONE[];
extern char str_AAM_RECOGNIZE_START[];
extern char str_AAM_RECOG_NORMALIZED[];
extern char str_AAM_MEET_NORMALIZED[];
extern char str_AAM_RECOG_CLASS[];
extern char str_AAM_RECOG_CLASS_DONE[];
extern char str_AAM_RECOGNIZE_DONE[];
extern char str_AAM_RECOGNIZE_PERCENT[];
extern char str_RECOGNITION_RESULT_Y[];
extern char str_RECOGNITION_RESULT_N[];
extern char str_AAM_F2CMP[];
extern char str_AAM_CMP2F[];
extern char str_AAM_COS_DONE[];

/* Messages */
extern char str_MSG_MORE_H[];
extern char str_MSG_MORE_M[];
extern char str_MSG_MORE_S[];
extern char str_MSG_MORE_MS[];

#endif /* STR_H */
