#include "stdafx.h"

#include "CAAM.h"
#include "CMatPtr.h"
#include "utilcpp.h"
#include "str.h"
#include "def.h"
#include "error.h"
#include "cvect.h"
#include "str2.h"

#include "lvq.h"

extern "C" {
#include <math.h>
#include <stdlib.h>
#include <ctype.h>
#include <stdio.h>
#include <errno.h>
}

#define INDETERMINATE		-5
void NChar(char *buf, int n);

CAAM::CAAM()
	{
	int i;
	
	pTrainingDB = NULL;
	pEFace = NULL;
	pEVect = NULL;
	pEVal = NULL;
	
	pReconstructionDB = NULL;
	pReconstructedDB = NULL;
	pReconstructedMat = NULL;
	pReconstructionCmp = NULL;
	
	pRecognizeDB = NULL;						
	pRecognizedDB = NULL;						
	pRecognizedMat = NULL;
	pRecognizeCmp = NULL;
	pMetDB = NULL;
	pMetCmp = NULL;
	pSammon = NULL;
	pCategorizeDB = NULL;
	
	for (i = 0; i < CMD_MAX; i++)
		bCommand[i] = FALSE;
	}

CAAM::~CAAM()
	{
	if (pSammon != NULL)
		delete pSammon;
		
	if (pEFace != NULL)
		delete pEFace;
		
	if (pEVal != NULL)
		delete pEVal;
	if (pEVect != NULL)
		{
		pEVect->FreeCol();
		delete pEVect;
		}
		
	if (pCategorizeDB != NULL)
		delete pCategorizeDB;
	if (pTrainingDB != NULL)
		delete pTrainingDB;
	if (pReconstructionDB != NULL)
		delete pReconstructionDB;
	if (pReconstructedDB != NULL)
		delete pReconstructedDB;
	if (pReconstructedMat != NULL)
		delete pReconstructedMat;
	if (pMetDB != NULL)
		delete pMetDB;
	if (pRecognizeDB != NULL)
		delete pRecognizeDB;
	if (pRecognizedDB != NULL)
		delete pRecognizedDB;
	if (pRecognizedMat != NULL)
		delete pRecognizedMat;
	if (pMetCmp != NULL)
		delete pMetCmp;
	if (pReconstructionCmp != NULL)
		delete pReconstructionCmp;
	if (pRecognizeCmp != NULL)
		delete pRecognizeCmp;
	}

/*****************************************************************************************
* Build the memory and compute its eigenvectors
* The Training set (pTrainingDB) has to be loaded before the calling to this fct.
*****************************************************************************************/
int CAAM::Build(long leigen)
	{
	long number; 
	long lFace = pTrainingDB->GetNumFace();
	long lPixel = pTrainingDB->GetNumPixel();
	CMatrixPtr *pX;											/*Matrix containing the training set, whose each columns points to a face data*/
	CMatrixPtr *pQ;
	
	Msg(1, str_AAM_BUILD_START);
	
	lRow = lPixel;
	lCol = lFace;
	if (leigen == -1)
		lEigen = lFace;
	else
		lEigen = lmin(leigen, lFace);
	
	pX = new CMatrixPtr(lRow, lCol);						/*Construct the matrix whose each columns points to a face data*/
	pQ = new CMatrixPtr(lCol, lCol);						/*Construct an intermediate matrix*/
	pQ->AllocateCol();										/*Allocate memory to store its columns*/
	pEVal = new CDiag(lCol);								/*Diagonal matrix containing the eigenvalues*/
	
	DBtoMat(pX, pTrainingDB);								/*Copy the pointer of the faces' data into the one of the columns' matrix*/
	pX->NormalizeCol();										/*Normalize each face*/
	Msg(0, str_AAM_BUILD_NORMALIZED);
	            
	pX->InnerProduct(pQ);									/*pQ = transpose(pX) * pX */
	Msg(0, str_AAM_BUILD_INNER);
	
	pQ->Eigen(pEVal, lEigen);								/*put in pQ its eigenvectors and in pEVal its eigenvalues*/
	Msg(0, str_AAM_BUILD_EIGEN1);
	
	Sort(pQ, pEVal);											/*Sort in descending order the eigenvalues in order to put*/
	Msg(0, str_AAM_BUILD_SORT);							/*eigenvectors whose have the biggest eigenvalues at the biginning of the matrix*/
	
	lEigen = lmin(lEigen, pQ->GetCol());            /*Sort may suppress columns, lEigen has to be les or equal than the number of col of pQ*/
	pEVect = new CMatrixPtr(lRow, lEigen);				/*Construct the eigenvector matrix*/
	pEVect->AllocateCol();									/*Allocate memory to store its columns*/
	
	if (leigen == -1 && number == 0)										
		pEVect->Multiply(pX, pQ);							/*pEVect = pX * pQ*/
	else
		pEVect->Multiply(pX, pQ, TRUE, lPixel, lEigen);		/*the same but this one is executed if the user asked for less eigenvectors*/
		
	pEVal->Pow(-0.5);										/*pEVal = PEVal ^ -1/2*/
	if (leigen == -1 && number == 0)
		pEVect->Multiply(pEVal);							/*pEVect = pEVect * pEVal*/
	else	
		pEVect->Multiply(pEVal, TRUE);						/*the same but this one is executed if the user asked for less eigenvectors*/
	pEVal->Pow(-2);											/*Reset the eigenvalues to their real values*/
	delete pQ;												/*delete the intermediate matrix*/
	delete pX;												/*delete the matrix storing the training set (but the set remains in the pTrainingDB object)*/
	GetEigenDB()->Copy(pEVect, pEVal);						/*Copy the pointer of each eigenvectors into the pointer of each face's data*/
	
	Msg(-1, str_AAM_BUILD_DONE, lEigen);
	return 0;
	}

/*****************************************************************************************
* Reconstruct the faces stored in pReconstructionDB
* pReconstructionDB has to be loaded before the calling to this fct.
*****************************************************************************************/
int CAAM::Reconstruct()
	{
	CMatrixPtr	*pReconstructionMat;						/*Matrix containing the pixel of the faces to be reconstructed*/
	CDiag		*pReconstructionCos;							/*Cosine between the real faces and the reconstructed ones*/
	long lRow, lCol;
	long lFace = pReconstructionDB->GetNumFace();
	long lPixel = pReconstructionDB->GetNumPixel();
	
	Msg(1, str_AAM_RECONSTRUCTION_START);
	
	lRow = lPixel;
	lCol = lFace;
	pReconstructionMat = new CMatrixPtr(lRow, lCol);
	pReconstructedMat = new CMatrixPtr(lRow, lCol);
	pReconstructedMat->AllocateCol();
	pReconstructionCmp = new CMatrix(lEigen, lCol);
	pReconstructionCos = new CDiag(lFace);
	pReconstructedDB = new CFaceDB();
	
	DBtoMat(pReconstructionMat, pReconstructionDB);	/*Copy the pointer of the faces' data into the one of the columns' matrix*/
	
	pReconstructionMat->NormalizeCol();					/*Normalize each face*/
	Msg(0, str_AAM_RECONS_NORMALIZED);
	
	FaceToCmp(pReconstructionCmp, pReconstructionMat);		
		/*Compute the composants of the unknown faces*/
	
	CmpToFace(pReconstructedMat, pReconstructionCmp);		
		/*Compute a set of faces from a set of composants*/
	
	ComputeCos(pReconstructionCos, pReconstructionMat, pReconstructedMat);
	
	pReconstructedDB->Copy(pReconstructedMat, pReconstructionCos);
		/*Copy the pointer of each eigenvectors into the pointer of each face's data*/
	
	delete pReconstructionMat;
	delete pReconstructionCos;
	Msg(-1, str_AAM_RECONSTRUCTION_DONE, lFace);
	return 0;
	}

/*****************************************************************************************
* Recognize the faces stored in pRecognizeDB.
* pRecognizeDB has to be loaded before the calling to this fct.
* It can, of course, recognize only the people previously seen. The composants of those people
* are stored in the pMetCmp matrix. If the face database pMetDB is null, it assume that those
* people are actually the people of the training set. So, in that case, this fct compute the
* composants of the training set at its beginning.
*****************************************************************************************/
int CAAM::Recognize(int nElect, BOOL bRecons)
	{
	CMatrixPtr	*pRecognizeMat = NULL;					/*Matrix containing the pixel of the faces to be recognized by th memory*/
	CMatrixPtr	*pRecognizedCmp = NULL;					/*Composants of faces of the already met people, each column correspond to a recognized face*/
	CDiag			*pRecognizeCos;							/*Cosine between the original face to be recognized and the face known recognized face*/
	CMatrix		*pAssociation = NULL;					/*At the end of the process pAssociation[n] will be equal to the number of the meeting face which looks the most like the nth face to recognize*/
	char Name1[64];
	char Name2[64];
	
	long lMetFace, lRecFace, lCorrect, l;
	long lPixel;
	CMatrixPtr *pX = NULL;                          /*Matrix containing the training set, whose each columns points to a face data*/
	float fPercent, fProb;											
	BOOL bMet = TRUE;
		
	Msg(1, str_AAM_RECOGNIZE_START);
	
	if (pMetDB == NULL)
		{															/*The met people are those in the training set*/
		pMetDB = pTrainingDB;
		bMet = FALSE;
		}
	
	lPixel  = pMetDB->GetNumPixel();
	lMetFace = pMetDB->GetNumFace();
	
	lRow = lPixel;
	lCol = lMetFace;
	pX = new CMatrixPtr(lRow, lCol);						/*Construct the matrix whose each columns points to a face data*/
	pMetCmp = new CMatrixPtr(lEigen, lCol);
	pMetCmp->AllocateCol();									/*Allocate memory to store its columns*/
	
	DBtoMat(pX, pMetDB);										/*Copy the pointer of the faces' data into the one of the columns' matrix*/
	pX->NormalizeCol();
	Msg(0, str_AAM_MEET_NORMALIZED);
	
	FaceToCmp(pMetCmp, pX);									/*Compute the composants of the training set*/
	delete pX;
	
	lRecFace = pRecognizeDB->GetNumFace();
	
	pRecognizeMat = new CMatrixPtr(lRow, lRecFace);
	pRecognizedMat = new CMatrixPtr(lRow, lRecFace);
	pRecognizedMat->AllocateCol();
	pRecognizeCmp = new CMatrix(lEigen, lRecFace);
	pRecognizedCmp = new CMatrixPtr(lEigen, lRecFace);
	pRecognizeCos = new CDiag(lRecFace);
	pRecognizedDB = new CFaceDB();
	
	if (strcmp(pMetDB->GetFaceName(1), "") != 0 && strcmp(pRecognizeDB->GetFaceName(1), "") != 0)
		pAssociation = new CMatrix(lRecFace, 2);
	else
		pAssociation = NULL;
	
	DBtoMat(pRecognizeMat, pRecognizeDB);
	pRecognizeMat->NormalizeCol();
	Msg(0, str_AAM_RECOG_NORMALIZED);
	
	FaceToCmp(pRecognizeCmp, pRecognizeMat);
	
	RecognizeCmp(pRecognizedCmp, pRecognizeCmp, pMetCmp, pAssociation, nElect, pMetDB);
	
	if (bRecons == TRUE)
		{
		CmpToFace(pRecognizedMat, pRecognizedCmp);
		
		ComputeCos(pRecognizeCos, pRecognizedMat, pRecognizeMat);
		
		pRecognizedDB->Copy(pRecognizedMat, pRecognizeCos);
		}
			
	delete pRecognizeMat;
	delete pRecognizedCmp;
	delete pRecognizeCos;
	if (pAssociation != NULL)
		{															/*If Face's name are given (by the user) for the met (or the training) database and for the faces to be recognized,*/
		lCorrect = 0;											/*Calculate the number of correct recognitions*/
		fProb = 0.0;
		pAssociation->Lock();
		for (l = 1; l <= lRecFace; l++)
			{
			if (pAssociation->GetAt(l, 1) == INDETERMINATE)
				{													/*Impossible to recognize the face*/
				Msg(0, "INDETERMINATE:   %s -> ???", pRecognizeDB->GetFaceName(l));
				}
			else
				{
				strcpy(Name1, pMetDB->GetFaceName(pAssociation->GetAt(l, 1)));
				strcpy(Name2, pRecognizeDB->GetFaceName(l));
				if (strcmp(Name1, Name2) == 0)
					{
					lCorrect++;
					Msg(0, str_RECOGNITION_RESULT_Y, Name2, Name1, pAssociation->GetAt(l, 2));
					fProb += pAssociation->GetAt(l, 2);
					}
				else
					Msg(0, str_RECOGNITION_RESULT_N, Name2, Name1, pAssociation->GetAt(l, 2));
				}
			}
		
		pAssociation->Unlock();
		fPercent = (float)lCorrect / (float)lRecFace;
		Msg(-1, str_AAM_RECOGNIZE_PERCENT, lRecFace, lCorrect, fPercent*100.0);
		Msg(0, "Average prob. of good recognition = %.1f", fProb/(float)lCorrect);
		}                                               
	else
		Msg(-1, str_AAM_RECOGNIZE_DONE, lRecFace);
	if (bMet == FALSE)
		{
		pMetDB = NULL;
		}
	return 0;
	}

/*****************************************************************************************
* Categorize the faces stored in pCategorizeDB. This categorization is done with the LVQ program
* pCategorizeDB has to be loaded before the calling to this fct.
* The learning stage of LVQ is achieved with the faces of pMetDB. If the face database pMetDB is 
* null, it assume that those faces are actually the faces of the training set. So, in that case, 
* this fct compute the composants of the training set at its beginning.
*****************************************************************************************/
int CAAM::Categorize(/*PFILECMD pcmd*/PFILESTATE fs)
	{
	CMatrixPtr	*pCategorizeMat = NULL;					/*Matrix containing the pixel of the faces to be recognized by th memory*/
	CMatrix	*pCategorizeCmp = NULL;						/*Composants of faces of the already met people, each column correspond to a recognized face*/
	
	long lMetFace, lCatFace, l;
	long lPixel;
	CMatrixPtr *pX = NULL;                          /*Matrix containing the training set, whose each columns points to a face data*/
	BOOL bMet = TRUE;
	float total;
		
	Msg(1, str_AAM_CATEGORIZE_START);
	
	if (pMetDB == NULL)
		{															/*The met people are those in the training set*/
		pMetDB = pTrainingDB;
		bMet = FALSE;
		}
	
	lPixel  = pMetDB->GetNumPixel();
	lMetFace = pMetDB->GetNumFace();
	
	lRow = lPixel;
	lCol = lMetFace;
	pX = new CMatrixPtr(lRow, lCol);						/*Construct the matrix whose each columns points to a face data*/
	pMetCmp = new CMatrixPtr(lEigen, lCol);
	pMetCmp->AllocateCol();									/*Allocate memory to store its columns*/
	
	DBtoMat(pX, pMetDB);										/*Copy the pointer of the faces' data into the one of the columns' matrix*/
	pX->NormalizeCol();
	Msg(0, str_AAM_MEET_NORMALIZED);
	
	FaceToCmp(pMetCmp, pX);									/*Compute the composants of the training set*/
	delete pX;
	
	lCatFace = pCategorizeDB->GetNumFace();
	
	pCategorizeMat = new CMatrixPtr(lRow, lCatFace);
	pCategorizeCmp = new CMatrix(lEigen, lCatFace);
	
	DBtoMat(pCategorizeMat, pCategorizeDB);
	pCategorizeMat->NormalizeCol();
	Msg(0, str_AAM_CATEG_NORMALIZED);
	
	FaceToCmp(pCategorizeCmp, pCategorizeMat);
   
   //put name into the matrix
   for (l = 1; l <= lMetFace; l++)
   	{
   	pMetCmp->SetColName(l, pMetDB->GetFaceName(l));
   	}
   
   for (l = 1; l <= lCatFace; l++)
   	{
   	pCategorizeCmp->SetColName(l, pCategorizeDB->GetFaceName(l));
   	}

#define UNDEF	0
#define INIT	1
#define LEARN	2
   
	//Categorization
	LVQPARAMS lvq;
	int nState = UNDEF;
	LVQInit(&lvq);
	BOOL bSammon;	
	lvq.pconfusion = NULL;
	if (/*pcmd == NULL*/fs->num_cmd == 0)
		{	/*default categorization*/
		lvq.noc 			= 9;
		lvq.knn 			= 1;
		lvq.xdim			= 3;
		lvq.ydim			= 3;
		lvq.neigh		= NEIGH_GAUSSIAN;
		lvq.topol		= TOPOL_RECT;
		lvq.randomize	= 10;
		lvq.alpha_type	= ALPHA_INVERSE_T;
		lvq.radius		= 90;
		lvq.alpha		= 0.3;
		lvq.pdata		= pMetCmp;
		lvq.pcode		= NULL;
		
		eveninit(&lvq);
		lvq.length = 100;
		olvq1(&lvq);
		lvq.length = 1000;
		lvq1(&lvq);
		lvq.pdata = pCategorizeCmp;
		lvq.pconfusion = NULL;
		accuracy(&lvq);
		}
	else
		{	/*user defined categorization*/
		lvq.pcode = NULL;
		bSammon = FALSE;
		
		int i = 0;
		
		while(i < fs->num_cmd /*pcmd != NULL*/)
			{
			/*first : parameters*/
			PFILECMD pcmd = &(fs->pnext[i]);
			i++;
			if (stricmp(pcmd->cmd, str_LVQ_NOC) == 0)
				{
				lvq.noc = atoi(pcmd->args);
				if (lvq.noc == 0)
					Msg(0, str_LVQ_ERR_PARAM, pcmd->args, str_LVQ_NOC);
				else
					Msg(0, "noc = %i", lvq.noc);
				}
			if (stricmp(pcmd->cmd, str_LVQ_KNN) == 0)
				{
				lvq.knn = atoi(pcmd->args);
				if (lvq.knn == 0)
					Msg(0, str_LVQ_ERR_PARAM, pcmd->args, str_LVQ_KNN);
				else
					Msg(0, "knn = %i", lvq.knn);
				}
			if (stricmp(pcmd->cmd, str_LVQ_XDIM) == 0)
				{
				lvq.xdim = atoi(pcmd->args);
				if (lvq.xdim == 0)
					Msg(0, str_LVQ_ERR_PARAM, pcmd->args, str_LVQ_XDIM);
				else
					Msg(0, "xdim = %i", lvq.xdim);
				}
			if (stricmp(pcmd->cmd, str_LVQ_YDIM) == 0)
				{
				lvq.ydim = atoi(pcmd->args);
				if (lvq.ydim == 0)
					Msg(0, str_LVQ_ERR_PARAM, pcmd->args, str_LVQ_YDIM);
				else
					Msg(0, "ydim = %i", lvq.ydim);
				}
			if (stricmp(pcmd->cmd, str_LVQ_NEIGH) == 0)
				{
				lvq.neigh = NEIGH_UNKNOWN;
				if (stricmp(pcmd->args, "bubble") == 0)
					{
					lvq.neigh = NEIGH_BUBBLE;
					Msg(0, "Neighbourhood type = Bubble");
					}
				if (stricmp(pcmd->args, "gaussian") == 0)
					{
					lvq.neigh = NEIGH_GAUSSIAN;
					Msg(0, "Neighbourhood type = Gaussian");
					}
				if (lvq.neigh == NEIGH_UNKNOWN)
					Msg(0, str_LVQ_ERR_PARAM, pcmd->args, str_LVQ_NEIGH);
				}
			if (stricmp(pcmd->cmd, str_LVQ_TOPOL) == 0)
				{
				lvq.topol = TOPOL_UNKNOWN;
				if (stricmp(pcmd->args, "data") == 0)
					{
					lvq.topol = TOPOL_DATA;
					Msg(0, "Topology type = Data");
					}
				if (stricmp(pcmd->args, "LVQ") == 0)
					{
					lvq.topol = TOPOL_LVQ;
					Msg(0, "Topology type = LVQ");
					}
				if (stricmp(pcmd->args, "HEXA") == 0)
					{
					lvq.topol = TOPOL_HEXA;
					Msg(0, "Topology type = Hexagonal");
					}
				if (stricmp(pcmd->args, "RECT") == 0)
					{
					lvq.topol = TOPOL_RECT;
					Msg(0, "Topology type = Rectangular");
					}
				if (lvq.topol == TOPOL_UNKNOWN)
					Msg(0, str_LVQ_ERR_PARAM, pcmd->args, str_LVQ_TOPOL);
				}
			if (stricmp(pcmd->cmd, str_LVQ_SEED) == 0)
				{
				lvq.randomize = atoi(pcmd->args);
				Msg(0, "seed = %i", lvq.randomize);
				}
			if (stricmp(pcmd->cmd, str_LVQ_ALPHA_TYPE) == 0)
				{
				lvq.alpha_type = ALPHA_UNKNOWN;
				if (stricmp(pcmd->args, "linear") == 0)
					{
					lvq.alpha_type = ALPHA_LINEAR;
					Msg(0, "Alpha decreasing = Linear");
					}
				if (stricmp(pcmd->args, "inverse_t") == 0)
					{
					lvq.alpha_type = ALPHA_INVERSE_T;
					Msg(0, "Alpha decreasing = Inverse Time");
					}
				if (lvq.alpha_type == ALPHA_UNKNOWN)
					Msg(0, str_LVQ_ERR_PARAM, pcmd->args, str_LVQ_ALPHA_TYPE);
				}
			if (stricmp(pcmd->cmd, str_LVQ_RADIUS) == 0)
				{
				lvq.radius = atof(pcmd->args);
				if (lvq.radius == 0.0)
					Msg(0, str_LVQ_ERR_PARAM, pcmd->args, str_LVQ_RADIUS);
				else
					Msg(0, "radius = %f", lvq.radius);
				}
			if (stricmp(pcmd->cmd, str_LVQ_ALPHA) == 0)
				{
				lvq.alpha = atof(pcmd->args);
				if (lvq.alpha == 0.0)
					Msg(0, str_LVQ_ERR_PARAM, pcmd->args, str_LVQ_ALPHA);
				else
					Msg(0, "initial alpha = %f", lvq.alpha);
				}
			if (stricmp(pcmd->cmd, str_LVQ_SAMMON) == 0)
				{
				if (lvq.sammon == NULL)
					{
					lvq.sammon = new CSammon();
					lvq.sammon->xdim = lvq.xdim;
					lvq.sammon->ydim = lvq.ydim;
					}

				char *str = strtok(pcmd->args, " \t");
				if (str == NULL)
					Msg(0, str_LVQ_ERR_PARAM, str, str_LVQ_SAMMON);
				lvq.sammon->times = atoi(str);
				if (lvq.sammon->times == 0)
					Msg(0, str_LVQ_ERR_PARAM, str, str_LVQ_SAMMON);
				
				str = strtok(NULL, " \t");
				if (str == NULL)
					Msg(0, str_LVQ_ERR_PARAM, pcmd->args, str_LVQ_SAMMON);
				lvq.sammon->length = atof(str);
				if (lvq.sammon->length == 0)
					Msg(0, str_LVQ_ERR_PARAM, str, str_LVQ_SAMMON);
				else
					{
					bSammon = TRUE;
					if (lvq.sammon->length > 1.0)
						Msg(0, "sammon times num = %d, run length = %.0f", lvq.sammon->times, 
							lvq.sammon->length);
					else
						Msg(0, "sammon times num = %d, max mapping error = %f", lvq.sammon->times, 
							lvq.sammon->length);
					}
				}
				
			/*second : commands*/
			if (stricmp(pcmd->cmd, str_LVQ_EVENINIT) == 0)
				{
				lvq.pdata = pMetCmp;
				Msg(0, "eveninit");
				nState = UNDEF;
				eveninit(&lvq);
				}
			if (stricmp(pcmd->cmd, str_LVQ_PROPINIT) == 0)
				{
				lvq.pdata = pMetCmp;
				Msg(0, "propinit");
				nState = UNDEF;
				propinit(&lvq);
				}
			if (stricmp(pcmd->cmd, str_LVQ_BALANCE) == 0)
				{
				lvq.pdata = pMetCmp;
				Msg(0, "balance");
				if (nState != UNDEF)
					{
					Msg(0, "You must call eveninit or propinit at the beginning of the classification process");
					break;
					}
				balance(&lvq);
				}
			if (stricmp(pcmd->cmd, str_LVQ_ACCURACY) == 0)
				{
				lvq.pdata = pCategorizeCmp;
				lvq.pconfusion = NULL;
				if (nState != LEARN)
					{
					Msg(0, "You must call one of the learning method (LVQ1, OLVQ1, LVQ2, or LVQ3) before accuracy");
					break;
					}
				Msg(0, "accuracy");
				accuracy(&lvq);
				}
			if (stricmp(pcmd->cmd, str_LVQ_SAVE) == 0)
				{
				if (lvq.pcode == NULL)
					{
					Msg(0, "Impossible to save the codebook vectors because they don't exist");
					}
				lvq.pcode->Save(pcmd->args);
				}
			if (stricmp(pcmd->cmd, str_LVQ_CLASSIFY) == 0)
				{
				lvq.pdata = pCategorizeCmp;
				lvq.pconfusion = NULL;
				Msg(0, "classify");
				if (nState != LEARN)
					{
					Msg(0, "You must call one of the learning method (LVQ1, OLVQ1, LVQ2, or LVQ3) before accuracy");
					break;
					}
				classify(&lvq);
				}
			if (stricmp(pcmd->cmd, str_LVQ_OLVQ1) == 0)
				{
				if (bSammon == FALSE && lvq.sammon != NULL)
					lvq.sammon->length = 0;
				lvq.length = atol(pcmd->args);
				if (lvq.length == 0L)
					Msg(0, str_LVQ_ERR_PARAM, pcmd->args, str_LVQ_OLVQ1);
				else
					{
					if (nState != UNDEF && nState != LEARN)
						{
						Msg(0, "You must call eveninit or propinit at the beginning of the classification process");
						break;
						}
					lvq.pdata = pMetCmp;
					Msg(0, "olvq1 iteration length = %.0f", (double)lvq.length);
					olvq1(&lvq);
					nState = LEARN;
					}
				bSammon = FALSE;
				}
			if (stricmp(pcmd->cmd, str_LVQ_LVQ1) == 0)
				{
				if (bSammon == FALSE && lvq.sammon != NULL)
					lvq.sammon->length = 0;
				lvq.length = atol(pcmd->args);
				if (lvq.length == 0L)
					Msg(0, str_LVQ_ERR_PARAM, pcmd->args, str_LVQ_LVQ1);
				else
					{
					if (nState != UNDEF && nState != LEARN)
						{
						Msg(0, "You must call eveninit or propinit at the beginning of the classification process");
						break;
						}
					lvq.pdata = pMetCmp;
					Msg(0, "lvq1 iteration length = %.0f", (double)lvq.length);
					lvq1(&lvq);
					nState = LEARN;
					}
				bSammon = FALSE;
				}
			if (stricmp(pcmd->cmd, str_LVQ_LVQ2) == 0)
				{
				if (bSammon == FALSE && lvq.sammon != NULL)
					lvq.sammon->length = 0;
				char *str = strtok(pcmd->args, " \t");
				if (str == NULL)
					Msg(0, str_LVQ_ERR_PARAM, pcmd->args, str_LVQ_LVQ2);
				lvq.length = atol(str);
				if (lvq.length == 0L)
					Msg(0, str_LVQ_ERR_PARAM, pcmd->args, str_LVQ_LVQ2);
				
				str = strtok(NULL, " \t");
				if (str == NULL)
					Msg(0, str_LVQ_ERR_PARAM, pcmd->args, str_LVQ_LVQ2);
				lvq.winlen = atof(str);
				if (lvq.winlen == 0.0)
					Msg(0, str_LVQ_ERR_PARAM, pcmd->args, str_LVQ_LVQ2);
				else
					{
					if (nState != UNDEF && nState != LEARN)
						{
						Msg(0, "You must call eveninit or propinit at the beginning of the classification process");
						break;
						}
					lvq.pdata = pMetCmp;
					Msg(0, "lvq2 iteration length = %.0f, window length = %.1f", (double)lvq.length, 
						lvq.winlen);
					lvq2(&lvq);
					nState = LEARN;
					}
				bSammon = FALSE;
				}

			if (stricmp(pcmd->cmd, str_LVQ_LVQ3) == 0)
				{
				if (bSammon == FALSE && lvq.sammon != NULL)
					lvq.sammon->length = 0;
				char *str = strtok(pcmd->args, " \t");
				if (str == NULL)
					Msg(0, str_LVQ_ERR_PARAM, pcmd->args, str_LVQ_LVQ3);
				lvq.length = atol(str);
				if (lvq.length == 0L)
					Msg(0, str_LVQ_ERR_PARAM, pcmd->args, str_LVQ_LVQ3);
				
				str = strtok(NULL, " \t");
				if (str == NULL)
					Msg(0, str_LVQ_ERR_PARAM, pcmd->args, str_LVQ_LVQ3);
				lvq.winlen = atof(str);
				if (lvq.winlen == 0.0)
					Msg(0, str_LVQ_ERR_PARAM, pcmd->args, str_LVQ_LVQ3);

				str = strtok(NULL, " \t");
				if (str == NULL)
					Msg(0, str_LVQ_ERR_PARAM, pcmd->args, str_LVQ_LVQ3);
				lvq.epsilon = atof(str);
				if (lvq.epsilon == 0.0)
					Msg(0, str_LVQ_ERR_PARAM, pcmd->args, str_LVQ_LVQ3);
				else
					{
					if (nState != UNDEF && nState != LEARN)
						{
						Msg(0, "You must call eveninit or propinit at the beginning of the classification process");
						break;
						}
					lvq.pdata = pMetCmp;    
					Msg(0, "lvq3 iteration length = %.0f, window length = %.1f, epsilon = %.1f", 
						(double)lvq.length, lvq.winlen, lvq.epsilon);
					lvq3(&lvq);
					nState = LEARN;
					}
				bSammon = FALSE;
				}
			
			pcmd = pcmd->pnext;
			}
		
		}
	
	char *str, buf[20];
	long i, i2, num_class;
	if (lvq.pconfusion != NULL)
		{
		Msg(0, "Accuracy results :");
		num_class = lvq.pconfusion->GetCol();
		str = (char*)malloc(num_class * sizeof(buf));
		if (str == NULL)
			goto end;
		
		strcpy(str, "     ");
		for (i = 0; i < num_class; i++)
			{
			_snprintf(buf, sizeof(buf)," %s", GetClassStr(i));
			NChar(buf, 7);
			strcat(str, buf);
			}
		Msg(0, str);
			
		lvq.pconfusion->Lock();
		for (i = 1; i <= num_class; i++)
			{
			_snprintf(buf, sizeof(buf), "%s", GetClassStr(i-1));
			NChar(buf, 7);
			strcpy(str, buf);
			total = 0.0;
			for (i2 = 1; i2 <= num_class; i2++)
				{
				total += lvq.pconfusion->GetAt(i, i2);
				}
			for (i2 = 1; i2 <= num_class && total != 0; i2++)
				{
				_snprintf(buf, sizeof(buf), "%.0f%%%% ", lvq.pconfusion->GetAt(i, i2)/total*100.0);
				NChar(buf, 7);
				strcat(str, buf);
				}
			Msg(0, str);
			}
		lvq.pconfusion->Unlock();

		free(str);
		}

end:	
	pSammon = lvq.sammon;
	delete lvq.pconfusion;
	LVQEnd(&lvq);
	delete pCategorizeMat;
	delete pCategorizeCmp;
	Msg(-1, str_AAM_CATEGORIZE_DONE, lCatFace);

	if (bMet == FALSE)
		{
		pMetDB = NULL;
		}
	
	return 0;
	}

void NChar(char *buf, int n)
	{
	int i;
	if (strlen(buf) > n)
		{
		buf[n] = 0;
		}
	else
		{
		i = strlen(buf);
		while (i < n)
			{
			buf[i] = ' ';
			buf[i+1] = 0;
			i++;
			}
		}
   }
   
CFaceDB *CAAM::GetCategorizeDB()
	{
	if (pCategorizeDB == NULL)
		pCategorizeDB = new CFaceDB();
		
	return pCategorizeDB;
	}

CFaceDB *CAAM::GetTrainingDB()
	{
	if (pTrainingDB == NULL)
		pTrainingDB = new CFaceDB();
		
	return pTrainingDB;
	}

CFaceDB *CAAM::GetEigenDB()
	{
	if (pEFace == NULL)
		pEFace = new CFaceDB();
		
	return pEFace;
	}

CFaceDB *CAAM::GetReconstructionDB()
	{
	if (pReconstructionDB == NULL)
		pReconstructionDB = new CFaceDB();
		
	return pReconstructionDB;
	}

CFaceDB *CAAM::GetRecognitionDB()
	{
	if (pRecognizeDB == NULL)
		pRecognizeDB = new CFaceDB();
		
	return pRecognizeDB;
	}

CFaceDB *CAAM::GetMetDB()
	{
	if (pMetDB == NULL)
		pMetDB = new CFaceDB();
		
	return pMetDB;
	}

CFaceDB *CAAM::GetReconstructedDB()
	{return pReconstructedDB;}

CFaceDB *CAAM::GetRecognizedDB()
	{return pRecognizedDB;}

CMatrix *CAAM::GetReconstructedMat()
	{return pReconstructedMat;}

CMatrix *CAAM::GetMetMat()
	{
	if (pMetMat == NULL)
		{
		pMetMat = new CMatrix();							/*Construct the eigenvector matrix*/
		}
		
	return pMetMat;
	}

CMatrixPtr *CAAM::GetEigenMat()
	{
	if (pEVect == NULL)
		{
		pEVect = new CMatrixPtr();							/*Construct the eigenvector matrix*/
		}
		
	return pEVect;
	}

CDiag *CAAM::GetEigenVal()
	{
	if (pEVal == NULL)
		{
		pEVal = new CDiag();
		}
		
	return pEVal;
	}

CMatrixPtr *CAAM::GetMeetCmp()
	{return pMetCmp;}

CMatrix *CAAM::GetReconstructionCmp()
	{return pReconstructionCmp;}

CMatrix *CAAM::GetRecognitionCmp()
	{return pRecognizeCmp;}

BOOL CAAM::IsCommand(int cmd)
	{
	if (cmd >= 0 && cmd < CMD_MAX)
		return bCommand[cmd];
	else
		return FALSE;
	}

/*****************************************************************************************
* Compute the composant of a set of face. The composants of the face n (in pFaceMat) is in 
* the column n (of pCmpMat). 
* Therefore pFaceMat must have the same number of columns than pCmpMat.
* The number of rows of pCmpMat must be the number of eigenvectors of the memory.
*****************************************************************************************/
void CAAM::FaceToCmp(CMatrix *pCmpMat, CMatrix *pFaceMat)
	{
	pEVect->SetTranspose(TRUE);
	pCmpMat->Multiply(pEVect, pFaceMat);
	pEVect->SetTranspose(FALSE);
	Msg(0, str_AAM_F2CMP);
	}

/*****************************************************************************************
* Compute a set of faces from a set of composants. 
* pFaceMat must have the same number of columns than pCmpMat, in other words the number of
* composant vectors is equal to the number of recalled faces.
* pCmpMat must have the same number of rows than the number of eigenvectors of the memory.
*****************************************************************************************/
void CAAM::CmpToFace(CMatrix *pFaceMat, CMatrix *pCmpMat)		
	{
	pFaceMat->Multiply(pEVect, pCmpMat);
	Msg(0, str_AAM_CMP2F);
	}
	
/*=======================================================================================*/

/*****************************************************************************************
* Copy the pointer of the faces' data into the one of the columns' matrix
******************************************************************************************/
void DBtoMat(CMatrixPtr *pMat, CFaceDB *pDB)
	{
	long l;
	long lCol = pDB->GetNumFace();
	
	for (l = 1; l <= lCol; l++)
		{													/*Copy the pointer of the faces' data into the one of the columns' matrix*/
		pMat->SetColumn(l, pDB->GetFaceMem(l));
		}
	}

void DBtoDiag(CDiag *pDiag, CFaceDB *pDB)
	{
	long l;
	long lCol = pDB->GetNumFace();
	
	pDiag->Lock();
	
	for (l = 1; l <= lCol; l++)
		{													
		pDiag->SetAt(l, atof(pDB->GetFaceName(l)));
		}
	pDiag->Unlock();
	}

/*****************************************************************************************
* Put in the diagonal matrix pCos, the cosine of the angle between the vectors of the
* matrices pMat1 and pMat2
* The Cosine of the angle between the column vector n of pMat1 (nth column of pMat1) and
* the column vector n of pMat2 (nth column of pMat2) is stored in the nth element of pCos 
*****************************************************************************************/
void ComputeCos(CDiag *pCos, CMatrix *pMat1, CMatrix *pMat2)
	{
	pMat1->SetTranspose(TRUE);
	pCos->Multiply(pMat1, pMat2);
	pMat1->SetTranspose(FALSE);
	Msg(0, str_AAM_COS_DONE);
	}

/*****************************************************************************************
* RecognizeCmp make an association between several column vectors of one matrix to other
* vectors of another matrix. It takes each column vector of pSourceCmp and associate it to
* the most similar vector of pKnownCmp. 
* The definition of "most similar vector" must be understood in a large way : the nElect vectors
* of the pKownCmp Matrix nearest (in term of euclidian distance) to a particular vector of 
* pSourceCmp are found. Then, those vectors vote. Their vote is their name found in the Face 
* Database pMetDB. The vector whose name win the election is associated to the vector of 
* pSourceCmp. This operation is done for every vector of pSourceCmp.
*
* 		CMatrixPtr *pDestCmp : 	This fct will copy the vector which is associated to the ith vector 
*										of pSourceCmp to the ith column of this matrix.
*		CMatrix *pSourceCmp	:	The source matrix, from which the column vectors which have to be
*										associated to one vector of the matrix pKonwnCmp, are extracted.
*		CMatrixPtr *pKnownCmp:	This matrix contains the column vectors which will be associated to.
*		CMatrix *pAssociation:	This is the output of this fct. This fct will store in the ith 
*										element of its first vector, the number of the column of the matrix 
*										pKnownCmp which is associated to the ith vector of pSourceCmp, and
*										in its second column the probability of the association.
*		int nElect				:	The number of electors the user wants to have.
*		CFaceDB* pMetDB		:	Used to know the name of each vector of pKnownCmp.
*****************************************************************************************/
void RecognizeCmp(CMatrixPtr *pDestCmp, CMatrix *pSourceCmp, CMatrixPtr *pKnownCmp, CMatrix *pAssociation, 
	int nElect, CFaceDB* pMetDB)
	{
	long lsource, lknown, *lrec, lcolsource, lcolknown, lrow, r;
	float d, *dmin;
	CMemory mem1, mem2, mem3;
	int i, j, nwin, *nvote;
	
	mem1.Allocate(nElect*sizeof(float));
	mem2.Allocate(nElect*sizeof(long));
	mem3.Allocate(nElect*sizeof(int));
	dmin = (float*)mem1.GetPtr();
	lrec = (long*)mem2.GetPtr();
	nvote = (int*)mem3.GetPtr();
	
	Msg(1, str_AAM_RECOG_CLASS);
	
	lcolknown =  pKnownCmp->GetCol();
	lcolsource  = pSourceCmp->GetCol();
	lrow = pSourceCmp->GetLi();
	
	pSourceCmp->Lock();
	pKnownCmp->Lock();
	if (pAssociation != NULL)
		pAssociation->Lock();
	
	for (lsource = 1; lsource <= lcolsource; lsource++)
		{
		for (i = 0; i < nElect; i++)
			{
			dmin[i] = MAXFLOAT;
			lrec[i] = 0;
			}
		
		for (lknown = 1; lknown <= lcolknown; lknown++)
			{
			d = 0.0;
			for (r = 1; r <= lrow; r++)
				{
				d += pow(pKnownCmp->GetAt(r, lknown) - pSourceCmp->GetAt(r, lsource), 2.0);
				}
			d = sqrt(d);
			
			if (d < dmin[nElect-1])
				{
				i = nElect-1;
				while (i >= 0 && d < dmin[i])          /*Look if lknown must be in the five elector*/
					{
					i--;
					}
				i++;
				if (i < nElect)
					{												/*lknown must be in the five elector, its place is i*/
					for (j = nElect-1; j > i; j--)
						{
						dmin[j] = dmin[j-1];
						lrec[j] = lrec[j-1];
						}
					dmin[i] = d;
					lrec[i] = lknown;
					}
				}
			}
			
		if (pAssociation != NULL)
			{														
			if (nElect > 1)
				{													/*Make the electors vote*/
				for (i = 0; i < nElect; i++)
					{    											/*Count the number of votes for each elector*/
					nvote[i] = 0;
					for (j = i; j < nElect; j++)
						{
						if (strcmp(pMetDB->GetFaceName(lrec[i]), pMetDB->GetFaceName(lrec[j])) == 0)
							nvote[i]++;
						}
					}
				nwin = 0;
				for (i = 1; i < nElect; i++)
					{
					if (nvote[nwin] < nvote[i])
						nwin = i;
					}
					
				if (nvote[nwin] > 1)
					{
					pAssociation->SetAt(lsource, 1, lrec[nwin]);
					pAssociation->SetAt(lsource, 2, (float)nvote[nwin]/float(nElect)*100.0);
					}
				else
					pAssociation->SetAt(lsource, 1, INDETERMINATE);             
					
				pDestCmp->SetColumn(lsource, pKnownCmp->GetColumn(lrec[nwin]));
				}
			else
				{													/*No need to vote*/
				pAssociation->SetAt(lsource, 1, lrec[0]);
				pAssociation->SetAt(lsource, 2, -1);
				pDestCmp->SetColumn(lsource, pKnownCmp->GetColumn(lrec[0]));
				}
			}
		}
	
	pSourceCmp->Unlock();
	pKnownCmp->Unlock();
	
	mem1.ReleasePtr();
	mem2.ReleasePtr();
	mem3.ReleasePtr();
	
	if (pAssociation != NULL)
		pAssociation->Unlock();
	Msg(-1, str_AAM_RECOG_CLASS_DONE);
	}

/*****************************************************************************************
* Sort the eigenvectors and the eigenvalues acording to the eigenvalue (descending order)
* This fct checks also that there doesn't exist any null eigenvalue, if it finds one, it
* discards the eigenvalue and the eigenvector;
*****************************************************************************************/
long Sort(CMatrixPtr *pMat, CVect *pVect)
	{
	long lcol, l, l2, maxl;
	float maxval, val;
	int number = 0;
	
	lcol = pVect->GetLi();
	
	pVect->Lock();
	
	for (l = 1L; l <= lcol; l++)
		{
		maxval = pVect->GetAt(l);
		
		if (maxval < 1E-7)									/*A null eigenvalue has been found (because*/
			{														/*There were two equal vectors in the training set*/
			number++;											/*So discard the eigenvector*/
			pVect->Discard(l);								/*and the eigenvalue*/
			pMat->Discard(l);
			l--;
			lcol--;
			}
		else
			{
			for (l2 = l+1L; l2 <= lcol; l2++)
				{
				val = pVect->GetAt(l2);
				if (val > maxval)
					{
					maxval = val;
					maxl = l2;
					}
				}
				
			if (maxval != pVect->GetAt(l))
				{	/*A bigger value has been found, the order must be changed*/
				pMat->SwapCol(maxl, l);
				pVect->SwapCol(maxl, l);
				}
			}
		}
	
	pVect->Unlock();
	
	return number;
	}
